/* eslint-disable */
// Below classes are for managing the canvas layout.
/*
    Potential values are separated with '|'
    All pixels are in core pixels.
    Supports multi touch with 2 fingers.
    This class uses only native events. There is: No timer for anything & no longpress for desktop & no doubleclick for touch screen & no wheel for touchscreen etc.

    Propagated events:  All events are propagated between "bound sections".
                        If 2 sections overlap but not bound, only top section gets the event.

    lowestPropagatedBoundSection:
            This property applies to bound sections.
            Bound sections are like layers. They overlap entirely.
            Every bound section should share same zIndex unless there is a very good reason for different zIndex values.
            To make a section "bound" to another, one can use "boundToSection" property (section name - string).
            When 2 or more sections overlap, the top section will get the event first.
            If a section handles the event and calls stopPropagating function, the sections after that section won't get the event.
            Example scenario:
                Event: drag
                Top section handles event.
                Middle section handles event then calls "stopPropagating" function. From this point, top section still handles the event until the end of the event.
                Bottom section doesn't get the event.

            If stopPropagating function is called, it is valid until the end of the event. So for events like dragging, calling
            the function only once is enough. When the event is lasted, lowestPropagatedBoundSection variable is set to null automatically.

    New section's options:
    name: 'tiles' | 'row headers' | 'column headers' | 'markers & cursor' | 'shapes' | 'scroll' etc.
    anchor: 'top left' | 'top right' | 'bottom left' | 'bottom right' (Examples: For row & column headers, anchor will be 'top left'; If we would want to draw something sticked to bottom, it would be 'bottom left' or 'bottom right').
        One can also anchor sections to other sections' edges.
        Order is important, first variables are related to top or bottom, next ones are related to left or right.
        Examples:
        1- [["column header", "bottom", "some section name", "top", "top"], "left"]
            ^ If "column header" section exists, its bottom will be section's top
            If "some section name" exists, its top will be section's top.
            If none of them exists, canvas's top will be used as anchor.
            Canvas's left will be used as horizontal anchor.
        2- [["column header", "bottom", "ruler", "bottom", "top"], ["row header", "right", "left"]]

    position: [0, 0] | [10, 50] | [x, y] // Related to anchor. Example 'bottom right': P(0, 0) is bottom right etc. myTopLeft is updated according to position and anchor.
    size: [100, 100] | [10, 20] | [maxX, maxY] // This doesn't restrict the drawable area, that is up to implementation. Size is not important for expanded directions. Expandable size is assigned after calculations.
    zIndex: Elements with highest zIndex will be drawn on top.
    expand: '' | 'right' | 'left' | 'top' | 'bottom' | 'left right top bottom' (any combination)
    interactable: true | false // If false, only general events will be fired (onDraw, newDocumentTopLeft, onResize). Example: editing mode, background drawings etc.
    drawingOrder: Sections with the same zIndex value are drawn according to their drawing order values.
        Section with the highest drawing order is drawn on top (for specific zIndex).
        So, in terms of being drawn on top, priority is: zIndex > drawingOrder.
    processingOrder:

    Processing order feature is tricky, let's say you want something like this:

    --------------------
    |     top bar      |
    --------------------
    --------- ----------
    | left  | | tiles  |
    | bar   | | area   |
    --------- ----------

    Top bar's height will be static most probably. It needs to be drawn first, so it can be expanded to right.
    If top bar is drawn after tiles area, since "tiles area" will most probably be expanded to all directions, it will leave no space for top bar.
    So, tiles area will be drawn last.
    For above situation, processing orders would be (with the same zIndex):
    * top bar -> 1
    * left bar -> 2
    * tiles area -> 3

    And "expand" properties would be like below:
        top bar: 'right' from position (0, 0)
        left bar: 'top bottom' from position (0, 200) -> So it will go up and find where top bar ends, then go down and find where canvas ends.
        tiles area: 'top bottom left right' from position like (300, 300) -> So it won't overlap with resulting positions of others (or it can stuck inside one of them).

    Expandable sections' dimensions are calculated according to other sections with the same zIndex.

    Below events trigger a redraw:
    * Adding a new section.
    * Click.
    * Double click.
    * Renewing all sections (optional redraw).
    * Requesting a redraw.

    Every section has a "section" property inside "sectionProperties".

    parentSectionName property (parameter of addSection): New section is added and its size and myTopLeft properties are mirrored from its parent section.
        All other properties and behaviours are the same with any section.


    Event handling:
        Mouse event combinations:
            mouse down + mouse up + click
            mouse down + mouse up + click + mouse down + mouse up + click + double click
            mouse move (if mouse is down, "draggingSomething" = true)
            mouse down + mouse move (dragging) + mouse up
            mouse wheel
            mouse enter
            mouse leave

        Touch event combinations:
            mouse down + mouse up + click
            mouse down + long press + mouse up
            mouse down + mouse move (it means dragging, "draggingSomething" = true) + mouse up // There is no "mouse move" event without dragging (for touch events)
            mouse down + multi touch start + multi touch move + multi touch end
*/
// This class will be used internally by CanvasSectionContainer.
var CanvasSectionObject = /** @class */ (function () {
    function CanvasSectionObject(options) {
        this.context = null;
        this.myTopLeft = null;
        this.documentTopLeft = null; // Document top left will be updated by container.
        this.containerObject = null;
        this.dpiScale = null;
        this.name = null;
        this.backgroundColor = null; // Defult is null (container's background color will be used).
        this.borderColor = null; // Default is null (no borders).
        this.boundToSection = null;
        this.anchor = new Array(0);
        this.position = new Array(0);
        this.size = new Array(0);
        this.expand = new Array(0);
        this.isLocated = false; // location and size of the section computed yet ?
        this.processingOrder = null;
        this.drawingOrder = null;
        this.zIndex = null;
        this.interactable = true;
        this.isAnimating = false;
        this.sectionProperties = {};
        this.name = options.name;
        this.backgroundColor = options.backgroundColor ? options.backgroundColor : null;
        this.borderColor = options.borderColor ? options.borderColor : null;
        this.anchor = typeof options.anchor === 'string' ? options.anchor.split(' ') : options.anchor;
        this.position = options.position;
        this.size = options.size;
        this.expand = options.expand.split(' ');
        this.processingOrder = options.processingOrder;
        this.drawingOrder = options.drawingOrder;
        this.zIndex = options.zIndex;
        this.interactable = options.interactable;
        this.sectionProperties = options.sectionProperties ? options.sectionProperties : {};
        this.onInitialize = options.onInitialize ? options.onInitialize : function () { };
        this.onMouseMove = options.onMouseMove ? options.onMouseMove : function () { };
        this.onMouseDown = options.onMouseDown ? options.onMouseDown : function () { };
        this.onMouseUp = options.onMouseUp ? options.onMouseUp : function () { };
        this.onMouseEnter = options.onMouseEnter ? options.onMouseEnter : function () { };
        this.onMouseLeave = options.onMouseLeave ? options.onMouseLeave : function () { };
        this.onClick = options.onClick ? options.onClick : function () { };
        this.onDoubleClick = options.onDoubleClick ? options.onDoubleClick : function () { };
        this.onContextMenu = options.onContextMenu ? options.onContextMenu : function () { };
        this.onMouseWheel = options.onMouseWheel ? options.onMouseWheel : function () { };
        this.onLongPress = options.onLongPress ? options.onLongPress : function () { };
        this.onMultiTouchStart = options.onMultiTouchStart ? options.onMultiTouchStart : function () { };
        this.onMultiTouchMove = options.onMultiTouchMove ? options.onMultiTouchMove : function () { };
        this.onMultiTouchEnd = options.onMultiTouchEnd ? options.onMultiTouchEnd : function () { };
        this.onResize = options.onResize ? options.onResize : function () { };
        this.onDraw = options.onDraw ? options.onDraw : function () { };
        this.onDrawArea = options.onDrawArea ? options.onDrawArea : function () { };
        this.onNewDocumentTopLeft = options.onNewDocumentTopLeft ? options.onNewDocumentTopLeft : function () { };
        this.onRemove = options.onRemove ? options.onRemove : function () { };
        this.onAnimationEnded = options.onAnimationEnded ? options.onAnimationEnded : function () { };
    }
    CanvasSectionObject.prototype.getTestDivContainer = function () {
        var element = document.getElementById('test-div-' + this.name);
        return element;
    };
    return CanvasSectionObject;
}());
var CanvasSectionContainer = /** @class */ (function () {
    function CanvasSectionContainer(canvasDOMElement) {
        /*
            All events will be cached by this class and propagated to sections.
            This class should work also mouse & touch enabled (at the same time) devices. Users should be able to use both.
        */
        this.dpiScale = window.devicePixelRatio;
        this.sections = new Array(0);
        this.documentTopLeft = [0, 0];
        this.documentBottomRight = [0, 0];
        this.positionOnMouseDown = null;
        this.positionOnMouseUp = null;
        this.positionOnClick = null;
        this.positionOnDoubleClick = null;
        this.mousePosition = null;
        this.dragDistance = null;
        this.draggingSomething = false; // This will be managed by container, used by sections.
        this.sectionOnMouseDown = null; // (Will contain section name) When dragging, user can leave section borders, dragging will continue. Target section will be informed.
        this.sectionUnderMouse = null; // For mouse enter & leave events.
        this.draggingTolerance = 5; // This is for only desktop, mobile browsers seem to distinguish dragging and clicking nicely.
        this.multiTouch = false;
        this.touchCenter = null;
        this.potentialLongPress = false;
        this.clearColor = 'white';
        this.touchEventInProgress = false; // This prevents multiple calling of mouse down and up events.
        this.testing = false; // If this set to true, container will create a div element for every section. So, cypress tests can find where to click etc.
        this.lowestPropagatedBoundSection = null; // Event propagating to bound sections. The first section which stops propagating and the sections those are on top of that section, get the event.
        this.scrollLineHeight = 30; // This will be overridden.
        this.mouseIsInside = false;
        this.inZoomAnimation = false;
        this.zoomChanged = false;
        // Below variables are related to animation feature.
        this.animatingSectionName = null; // The section that called startAnimating function. This variable is null when animations are not running.
        this.lastFrameStamp = null;
        this.continueAnimating = null;
        this.frameCount = null; // Frame count of the current animation.
        this.duration = null; // Duration for the animation.
        this.elapsedTime = null; // Time that passed since the animation started.
        this.canvas = canvasDOMElement;
        this.context = canvasDOMElement.getContext('2d');
        this.context.setTransform(1, 0, 0, 1, 0, 0);
        document.addEventListener('mousemove', this.onMouseMove.bind(this));
        this.canvas.onmousedown = this.onMouseDown.bind(this);
        document.addEventListener('mouseup', this.onMouseUp.bind(this));
        this.canvas.onclick = this.onClick.bind(this);
        this.canvas.ondblclick = this.onDoubleClick.bind(this);
        this.canvas.oncontextmenu = this.onContextMenu.bind(this);
        this.canvas.onwheel = this.onMouseWheel.bind(this);
        this.canvas.onmouseleave = this.onMouseLeave.bind(this);
        this.canvas.onmouseenter = this.onMouseEnter.bind(this);
        this.canvas.ontouchstart = this.onTouchStart.bind(this);
        this.canvas.ontouchmove = this.onTouchMove.bind(this);
        this.canvas.ontouchend = this.onTouchEnd.bind(this);
        this.canvas.ontouchcancel = this.onTouchCancel.bind(this);
        // Some explanation first.
        // When the user uses the mouse wheel for scrolling, different browsers use different technics for calculating the deltaY and deltaX values.
        // For example FireFox uses "deltaMode=1" which corresponds to "lines". So it creates the event with the number of lines to scroll.
        // Chrome uses "deltaMode=0" which corresponds to "pixels". So it creates the event with the number of pixels to scroll.
        // When "deltaMode=1" is used, we need to know the height of the line, so we will convert it to pixels.
        // For that purpose, we'll create a temporary div element, get the font size and delete the temporary element.
        var tempElement = document.createElement('div');
        tempElement.style.fontSize = 'initial'; // IE doesn't support this property, but it uses "deltaMode=0" (so we don't need to get the line height).
        tempElement.style.display = 'none';
        document.body.appendChild(tempElement);
        this.scrollLineHeight = parseInt(window.getComputedStyle(tempElement).fontSize);
        document.body.removeChild(tempElement); // Remove the temporary element.
    }
    CanvasSectionContainer.prototype.getContext = function () {
        return this.context;
    };
    CanvasSectionContainer.prototype.setClearColor = function (color) {
        this.clearColor = color;
    };
    CanvasSectionContainer.prototype.getClearColor = function () {
        return this.clearColor;
    };
    CanvasSectionContainer.prototype.setInZoomAnimation = function (inZoomAnimation) {
        this.inZoomAnimation = inZoomAnimation;
    };
    CanvasSectionContainer.prototype.isInZoomAnimation = function () {
        return this.inZoomAnimation;
    };
    CanvasSectionContainer.prototype.setZoomChanged = function (zoomChanged) {
        this.zoomChanged = zoomChanged;
    };
    CanvasSectionContainer.prototype.isZoomChanged = function () {
        return this.zoomChanged;
    };
    /**
     * IE11 doesn't support Array.includes, use replacement
    */
    CanvasSectionContainer.prototype.arrayIncludes = function (array, element) {
        return array.indexOf(element) >= 0;
    };
    ;
    CanvasSectionContainer.prototype.clearMousePositions = function () {
        this.positionOnClick = this.positionOnDoubleClick = this.positionOnMouseDown = this.positionOnMouseUp = this.dragDistance = this.sectionOnMouseDown = null;
        this.touchCenter = null;
        this.draggingSomething = false;
        this.touchEventInProgress = false;
        this.lowestPropagatedBoundSection = null;
    };
    CanvasSectionContainer.prototype.convertPositionToSectionLocale = function (section, point) {
        return [point[0] - section.myTopLeft[0], point[1] - section.myTopLeft[1]];
    };
    CanvasSectionContainer.prototype.convertPositionToCanvasLocale = function (e) {
        var rect = this.canvas.getBoundingClientRect();
        var x, y;
        if (e.touches !== undefined && e.touches.length > 0) {
            x = e.touches[0].clientX - rect.left;
            y = e.touches[0].clientY - rect.top;
        }
        else if (e.changedTouches !== undefined && e.changedTouches.length > 0) {
            x = e.changedTouches[0].clientX - rect.left;
            y = e.changedTouches[0].clientY - rect.top;
        }
        else {
            x = e.clientX - rect.left;
            y = e.clientY - rect.top;
        }
        return [Math.round(x * this.dpiScale), Math.round(y * this.dpiScale)];
    };
    CanvasSectionContainer.prototype.convertPointToCanvasLocale = function (point) {
        var rect = this.canvas.getBoundingClientRect();
        var x, y;
        x = point[0] - rect.left;
        y = point[1] - rect.top;
        return [Math.round(x * this.dpiScale), Math.round(y * this.dpiScale)];
    };
    CanvasSectionContainer.prototype.getSectionWithName = function (name) {
        if (name) {
            for (var i = 0; i < this.sections.length; i++) {
                if (this.sections[i].name === name) {
                    return this.sections[i];
                }
            }
            return null;
        }
        else {
            return null;
        }
    };
    CanvasSectionContainer.prototype.getDocumentTopLeft = function () {
        return [this.documentTopLeft[0], this.documentTopLeft[1]];
    };
    CanvasSectionContainer.prototype.setDocumentBounds = function (points) {
        this.documentTopLeft[0] = Math.round(points[0]);
        this.documentTopLeft[1] = Math.round(points[1]);
        this.documentBottomRight[0] = Math.round(points[2]);
        this.documentBottomRight[1] = Math.round(points[3]);
        for (var i = 0; i < this.sections.length; i++) {
            this.sections[i].onNewDocumentTopLeft(this.getDocumentTopLeft());
        }
    };
    CanvasSectionContainer.prototype.updateBoundSectionList = function (section, sectionList) {
        if (sectionList === void 0) { sectionList = null; }
        if (sectionList === null)
            sectionList = new Array(0);
        sectionList.push(section);
        var tempSectionList = new Array(0);
        if (section.boundToSection) {
            var tempSection = this.getSectionWithName(section.boundToSection);
            if (tempSection && tempSection.isLocated) {
                if (!this.arrayIncludes(sectionList, tempSection))
                    tempSectionList.push(tempSection);
            }
        }
        for (var i = 0; i < this.sections.length; i++) {
            if (this.sections[i].isLocated && this.sections[i].boundToSection === section.name) {
                if (!this.arrayIncludes(sectionList, this.sections[i]))
                    tempSectionList.push(this.sections[i]);
            }
        }
        for (var i = 0; i < tempSectionList.length; i++) {
            this.updateBoundSectionList(tempSectionList[i], sectionList);
        }
        return sectionList;
    };
    CanvasSectionContainer.prototype.orderBoundsList = function (section) {
        // According to zIndex & drawingOrder.
        for (var i = 0; i < section.boundsList.length - 1; i++) {
            for (var j = i + 1; j < section.boundsList.length; j++) {
                if (section.boundsList[i].zIndex > section.boundsList[j].zIndex
                    || (section.boundsList[i].zIndex === section.boundsList[j].zIndex && section.boundsList[i].drawingOrder > section.boundsList[j].drawingOrder)) {
                    var temp = section.boundsList[i];
                    section.boundsList[i] = section.boundsList[j];
                    section.boundsList[j] = temp;
                }
            }
        }
        // Remove the sections those are above this section. Events will not be propagated to them.
        for (var i = section.boundsList.length - 1; i > -1; i--) {
            if (section.boundsList[i].name !== section.name) {
                section.boundsList.splice(i, 1);
            }
            else {
                break;
            }
        }
    };
    CanvasSectionContainer.prototype.updateBoundSectionLists = function () {
        for (var i = 0; i < this.sections.length; i++) {
            this.sections[i].boundsList = null;
            this.sections[i].boundsList = this.updateBoundSectionList(this.sections[i]);
            this.orderBoundsList(this.sections[i]);
        }
    };
    CanvasSectionContainer.prototype.requestReDraw = function () {
        if (!this.getAnimatingSectionName())
            this.drawSections();
    };
    CanvasSectionContainer.prototype.propagateOnClick = function (section, position, e) {
        for (var i = section.boundsList.length - 1; i > -1; i--) {
            if (section.boundsList[i].interactable)
                section.boundsList[i].onClick(position, e);
            if (section.boundsList[i].name === this.lowestPropagatedBoundSection)
                break; // Stop propagation.
        }
    };
    CanvasSectionContainer.prototype.propagateOnDoubleClick = function (section, position, e) {
        for (var i = section.boundsList.length - 1; i > -1; i--) {
            if (section.boundsList[i].interactable)
                section.boundsList[i].onDoubleClick(position, e);
            if (section.boundsList[i].name === this.lowestPropagatedBoundSection)
                break; // Stop propagation.
        }
    };
    CanvasSectionContainer.prototype.propagateOnMouseLeave = function (section, position, e) {
        for (var i = section.boundsList.length - 1; i > -1; i--) {
            if (section.boundsList[i].interactable)
                section.boundsList[i].onMouseLeave(position, e);
            if (section.boundsList[i].name === this.lowestPropagatedBoundSection)
                break; // Stop propagation.
        }
    };
    CanvasSectionContainer.prototype.propagateOnMouseEnter = function (section, position, e) {
        for (var i = section.boundsList.length - 1; i > -1; i--) {
            if (section.boundsList[i].interactable)
                section.boundsList[i].onMouseEnter(position, e);
            if (section.boundsList[i].name === this.lowestPropagatedBoundSection)
                break; // Stop propagation.
        }
    };
    CanvasSectionContainer.prototype.propagateOnMouseMove = function (section, position, dragDistance, e) {
        for (var i = section.boundsList.length - 1; i > -1; i--) {
            if (section.boundsList[i].interactable)
                section.boundsList[i].onMouseMove(position, dragDistance, e);
            if (section.boundsList[i].name === this.lowestPropagatedBoundSection)
                break; // Stop propagation.
        }
    };
    CanvasSectionContainer.prototype.propagateOnLongPress = function (section, position, e) {
        for (var i = section.boundsList.length - 1; i > -1; i--) {
            if (section.boundsList[i].interactable)
                section.boundsList[i].onLongPress(position, e);
            if (section.boundsList[i].name === this.lowestPropagatedBoundSection)
                break; // Stop propagation.
        }
    };
    CanvasSectionContainer.prototype.propagateOnMouseDown = function (section, position, e) {
        for (var i = section.boundsList.length - 1; i > -1; i--) {
            if (section.boundsList[i].interactable)
                section.boundsList[i].onMouseDown(position, e);
            if (section.boundsList[i].name === this.lowestPropagatedBoundSection)
                break; // Stop propagation.
        }
    };
    CanvasSectionContainer.prototype.propagateOnMouseUp = function (section, position, e) {
        for (var i = section.boundsList.length - 1; i > -1; i--) {
            if (section.boundsList[i].interactable)
                section.boundsList[i].onMouseUp(position, e);
            if (section.boundsList[i].name === this.lowestPropagatedBoundSection)
                break; // Stop propagation.
        }
    };
    CanvasSectionContainer.prototype.propagateOnContextMenu = function (section) {
        for (var i = section.boundsList.length - 1; i > -1; i--) {
            if (section.boundsList[i].interactable)
                section.boundsList[i].onContextMenu();
            if (section.boundsList[i].name === this.lowestPropagatedBoundSection)
                break; // Stop propagation.
        }
    };
    CanvasSectionContainer.prototype.propagateOnMouseWheel = function (section, position, delta, e) {
        for (var i = section.boundsList.length - 1; i > -1; i--) {
            if (section.boundsList[i].interactable)
                section.boundsList[i].onMouseWheel(position, delta, e);
            if (section.boundsList[i].name === this.lowestPropagatedBoundSection)
                break; // Stop propagation.
        }
    };
    CanvasSectionContainer.prototype.propagateOnMultiTouchStart = function (section, e) {
        for (var i = section.boundsList.length - 1; i > -1; i--) {
            if (section.boundsList[i].interactable)
                section.boundsList[i].onMultiTouchStart(e);
            if (section.boundsList[i].name === this.lowestPropagatedBoundSection)
                break; // Stop propagation.
        }
    };
    CanvasSectionContainer.prototype.propagateOnMultiTouchMove = function (section, position, distance, e) {
        for (var i = section.boundsList.length - 1; i > -1; i--) {
            if (section.boundsList[i].interactable)
                section.boundsList[i].onMultiTouchMove(position, distance, e);
            if (section.boundsList[i].name === this.lowestPropagatedBoundSection)
                break; // Stop propagation.
        }
    };
    CanvasSectionContainer.prototype.propagateOnMultiTouchEnd = function (section, e) {
        for (var i = section.boundsList.length - 1; i > -1; i--) {
            if (section.boundsList[i].interactable)
                section.boundsList[i].onMultiTouchEnd(e);
            if (section.boundsList[i].name === this.lowestPropagatedBoundSection)
                break; // Stop propagation.
        }
    };
    CanvasSectionContainer.prototype.onClick = function (e) {
        if (!this.draggingSomething) { // Prevent click event after dragging.
            if (this.positionOnMouseDown !== null && this.positionOnMouseUp !== null) {
                this.positionOnClick = this.convertPositionToCanvasLocale(e);
                var s1 = this.findSectionContainingPoint(this.positionOnMouseDown);
                var s2 = this.findSectionContainingPoint(this.positionOnMouseUp);
                if (s1 && s2 && s1 == s2) { // Allow click event if only mouse was above same section while clicking.
                    var section = this.findSectionContainingPoint(this.positionOnClick);
                    if (section) { // "interactable" property is also checked inside function "findSectionContainingPoint".
                        this.propagateOnClick(section, this.convertPositionToSectionLocale(section, this.positionOnClick), e);
                    }
                }
            }
            this.clearMousePositions(); // Drawing takes place after cleaning mouse positions. Sections should overcome this evil.
            this.drawSections();
        }
        else {
            this.clearMousePositions();
        }
    };
    CanvasSectionContainer.prototype.onDoubleClick = function (e) {
        this.positionOnDoubleClick = this.convertPositionToCanvasLocale(e);
        var section = this.findSectionContainingPoint(this.positionOnDoubleClick);
        if (section) {
            this.propagateOnDoubleClick(section, this.convertPositionToSectionLocale(section, this.positionOnDoubleClick), e);
        }
        this.clearMousePositions();
        this.drawSections();
    };
    CanvasSectionContainer.prototype.onMouseMove = function (e) {
        // Early exit. If mouse is outside and "draggingSomething = false", then there is no reason to check further.
        if (!this.mouseIsInside && !this.draggingSomething)
            return;
        if (!this.potentialLongPress) {
            if (!this.touchEventInProgress) {
                this.mousePosition = this.convertPositionToCanvasLocale(e);
                if (this.positionOnMouseDown !== null && !this.draggingSomething) {
                    var dragDistance = [this.mousePosition[0] - this.positionOnMouseDown[0], this.mousePosition[1] - this.positionOnMouseDown[1]];
                    if (Math.abs(dragDistance[0]) >= this.draggingTolerance || Math.abs(dragDistance[1]) >= this.draggingTolerance) {
                        this.draggingSomething = true;
                    }
                }
                var section;
                if (this.draggingSomething) {
                    this.dragDistance = [this.mousePosition[0] - this.positionOnMouseDown[0], this.mousePosition[1] - this.positionOnMouseDown[1]];
                    section = this.getSectionWithName(this.sectionOnMouseDown);
                }
                else {
                    section = this.findSectionContainingPoint(this.mousePosition);
                }
                if (section) {
                    if (section.name !== this.sectionUnderMouse) {
                        if (this.sectionUnderMouse !== null) {
                            var previousSection = this.getSectionWithName(this.sectionUnderMouse);
                            if (previousSection)
                                this.propagateOnMouseLeave(previousSection, this.convertPositionToSectionLocale(previousSection, this.mousePosition), e);
                        }
                        this.sectionUnderMouse = section.name;
                        this.propagateOnMouseEnter(section, this.convertPositionToSectionLocale(section, this.mousePosition), e);
                    }
                    this.propagateOnMouseMove(section, this.convertPositionToSectionLocale(section, this.mousePosition), this.dragDistance, e);
                }
                else if (this.sectionUnderMouse !== null) {
                    var previousSection = this.getSectionWithName(this.sectionUnderMouse);
                    if (previousSection)
                        this.propagateOnMouseLeave(previousSection, this.convertPositionToSectionLocale(previousSection, this.mousePosition), e);
                    this.sectionUnderMouse = null;
                }
            }
        }
        else {
            this.mousePosition = this.convertPositionToCanvasLocale(e);
            var section = this.findSectionContainingPoint(this.mousePosition);
            if (section) {
                this.propagateOnLongPress(section, this.convertPositionToSectionLocale(section, this.mousePosition), e);
            }
        }
    };
    CanvasSectionContainer.prototype.onMouseDown = function (e) {
        if (e.button === 0 && !this.touchEventInProgress) { // So, we only handle left button.
            this.clearMousePositions();
            this.positionOnMouseDown = this.convertPositionToCanvasLocale(e);
            var section = this.findSectionContainingPoint(this.positionOnMouseDown);
            if (section) {
                this.sectionOnMouseDown = section.name;
                this.propagateOnMouseDown(section, this.convertPositionToSectionLocale(section, this.positionOnMouseDown), e);
            }
        }
    };
    CanvasSectionContainer.prototype.onMouseUp = function (e) {
        // Early exit. If mouse down position is not inside the canvas area, we have nothing to check further.
        if (!this.positionOnMouseDown) {
            this.clearMousePositions();
            return;
        }
        if (e.button === 0 && !this.touchEventInProgress) {
            this.positionOnMouseUp = this.convertPositionToCanvasLocale(e);
            if (!this.draggingSomething) {
                var section = this.findSectionContainingPoint(this.positionOnMouseUp);
                if (section) {
                    this.propagateOnMouseUp(section, this.convertPositionToSectionLocale(section, this.positionOnMouseUp), e);
                }
            }
            else {
                var section = this.getSectionWithName(this.sectionOnMouseDown);
                if (section) {
                    this.propagateOnMouseUp(section, this.convertPositionToSectionLocale(section, this.positionOnMouseUp), e);
                }
            }
        }
        if (!this.mouseIsInside) { // Normally, onclick event clears the positions. In this case, onClick won't be fired. So we clear the positions.
            this.clearMousePositions();
        }
    };
    CanvasSectionContainer.prototype.onContextMenu = function (e) {
        var mousePosition = this.convertPositionToCanvasLocale(e);
        var section = this.findSectionContainingPoint(mousePosition);
        if (section) {
            this.propagateOnContextMenu(section);
        }
        if (this.potentialLongPress) {
            // LongPress triggers context menu.
            // We should stop propagating here because we are using different context menu handlers for touch and mouse events.
            // By stopping this event here, we can have real context menus (for mice) and other handlers (for longpress) at the same time (see Control.RowHeader.js).
            e.preventDefault();
            e.stopPropagation();
            return false;
        }
    };
    CanvasSectionContainer.prototype.onMouseWheel = function (e) {
        var point = this.convertPositionToCanvasLocale(e);
        var delta;
        if (e.deltaMode === 1)
            delta = [e.deltaX * this.scrollLineHeight, e.deltaY * this.scrollLineHeight];
        else
            delta = [e.deltaX, e.deltaY];
        var section = this.findSectionContainingPoint(point);
        if (section)
            this.propagateOnMouseWheel(section, this.convertPositionToSectionLocale(section, point), delta, e);
    };
    CanvasSectionContainer.prototype.onMouseLeave = function (e) {
        // While dragging something, we don't clear the event information even if the mouse is outside of the canvas area.
        // We catch the mouse move and mouse up events even when the mouse pointer is outside the canvas area.
        // This feature is enabled to create a better dragging experience.
        if (!this.draggingSomething) {
            if (this.sectionUnderMouse !== null) {
                var section = this.getSectionWithName(this.sectionUnderMouse);
                if (section)
                    this.propagateOnMouseLeave(section, null, e);
                this.sectionUnderMouse = null;
            }
            this.clearMousePositions();
            this.mousePosition = null;
        }
        this.mouseIsInside = false;
    };
    CanvasSectionContainer.prototype.onMouseEnter = function (e) {
        this.mouseIsInside = true;
    };
    CanvasSectionContainer.prototype.onTouchStart = function (e) {
        if (e.touches.length === 1) {
            this.clearMousePositions();
            this.potentialLongPress = true;
            this.positionOnMouseDown = this.convertPositionToCanvasLocale(e);
            var section = this.findSectionContainingPoint(this.positionOnMouseDown);
            if (section) {
                this.sectionOnMouseDown = section.name;
                this.propagateOnMouseDown(section, this.convertPositionToSectionLocale(section, this.positionOnMouseDown), e);
            }
        }
        else if (!this.multiTouch) {
            this.potentialLongPress = false;
            this.multiTouch = true;
            var section = this.getSectionWithName(this.sectionOnMouseDown);
            if (section)
                this.propagateOnMultiTouchStart(section, e);
        }
    };
    CanvasSectionContainer.prototype.onTouchMove = function (e) {
        // Sometimes onTouchStart is fired for another element. In this case, we return.
        if (this.positionOnMouseDown === null)
            return;
        this.potentialLongPress = false;
        if (!this.multiTouch) {
            this.mousePosition = this.convertPositionToCanvasLocale(e);
            this.draggingSomething = true;
            if (this.positionOnMouseDown !== null) {
                this.dragDistance = [this.mousePosition[0] - this.positionOnMouseDown[0], this.mousePosition[1] - this.positionOnMouseDown[1]];
                var section = this.getSectionWithName(this.sectionOnMouseDown);
                if (section) {
                    this.propagateOnMouseMove(section, this.convertPositionToSectionLocale(section, this.mousePosition), this.dragDistance, e);
                }
            }
        }
        else if (e.touches.length === 2) {
            var section = this.getSectionWithName(this.sectionOnMouseDown);
            if (section) {
                var diffX = Math.abs(e.touches[0].clientX - e.touches[1].clientX);
                var diffY = Math.abs(e.touches[0].clientY - e.touches[1].clientY);
                // Let's keep "touchCenter" variable "static" for now. When we want to allow move & drag at the same time, we should make it dynamic again.
                if (!this.touchCenter) {
                    this.touchCenter = [(e.touches[0].clientX + e.touches[1].clientX) * 0.5, (e.touches[0].clientY + e.touches[1].clientY) * 0.5];
                    this.touchCenter = this.convertPointToCanvasLocale(this.touchCenter);
                }
                var distance = Math.sqrt(Math.pow(diffX, 2) + Math.pow(diffY, 2));
                this.propagateOnMultiTouchMove(section, this.convertPositionToSectionLocale(section, this.touchCenter), distance, e);
            }
        }
    };
    CanvasSectionContainer.prototype.onTouchEnd = function (e) {
        this.potentialLongPress = false;
        if (!this.multiTouch) {
            this.positionOnMouseUp = this.convertPositionToCanvasLocale(e);
            if (!this.draggingSomething) {
                var section = this.findSectionContainingPoint(this.positionOnMouseUp);
                if (section)
                    this.propagateOnMouseUp(section, this.convertPositionToSectionLocale(section, this.positionOnMouseUp), e);
            }
            else {
                var section = this.getSectionWithName(this.sectionOnMouseDown);
                if (section)
                    this.propagateOnMouseUp(section, this.convertPositionToSectionLocale(section, this.positionOnMouseUp), e);
            }
        }
        else {
            if (e.touches.length === 0) {
                this.multiTouch = false;
                var section = this.getSectionWithName(this.sectionOnMouseDown);
                if (section) {
                    this.propagateOnMultiTouchEnd(section, e);
                }
            }
        }
        this.touchEventInProgress = true;
    };
    CanvasSectionContainer.prototype.onTouchCancel = function (e) {
        this.clearMousePositions();
        this.potentialLongPress = false;
    };
    CanvasSectionContainer.prototype.onResize = function (newWidth, newHeight) {
        var container = this.canvas.parentNode;
        var cRect = container.getBoundingClientRect();
        if (!newWidth)
            newWidth = cRect.right - cRect.left;
        if (!newHeight)
            newHeight = cRect.bottom - cRect.top;
        this.dpiScale = window.devicePixelRatio;
        newWidth = Math.floor(newWidth * this.dpiScale);
        newHeight = Math.floor(newHeight * this.dpiScale);
        this.canvas.width = newWidth;
        this.canvas.height = newHeight;
        // CSS pixels can be fractional, but need to round to the same real pixels
        var cssWidth = newWidth / this.dpiScale; // NB. beware
        var cssHeight = newHeight / this.dpiScale;
        this.canvas.style.width = cssWidth.toFixed(4) + 'px';
        this.canvas.style.height = cssHeight.toFixed(4) + 'px';
        this.clearMousePositions();
        this.right = this.canvas.width;
        this.bottom = this.canvas.height;
        for (var i = 0; i < this.sections.length; i++) {
            this.sections[i].dpiScale = this.dpiScale;
        }
        this.reNewAllSections();
    };
    CanvasSectionContainer.prototype.findSectionContainingPoint = function (point) {
        for (var i = this.sections.length - 1; i > -1; i--) { // Search from top to bottom. Top section will be sent as target section.
            if (this.sections[i].isLocated && this.sections[i].interactable && this.doesSectionIncludePoint(this.sections[i], point))
                return this.sections[i];
        }
        return null;
    };
    CanvasSectionContainer.prototype.doesSectionIncludePoint = function (section, point) {
        return ((point[0] >= section.myTopLeft[0] && point[0] <= section.myTopLeft[0] + section.size[0]) && (point[1] >= section.myTopLeft[1] && point[1] <= section.myTopLeft[1] + section.size[1]));
    };
    CanvasSectionContainer.prototype.doSectionsIntersectOnYAxis = function (section1, section2) {
        var y11 = section1.myTopLeft[1];
        var y12 = section1.myTopLeft[1] + section1.size[1];
        var y21 = section2.myTopLeft[1];
        var y22 = section2.myTopLeft[1] + section2.size[1];
        if (((y11 >= y21 && y11 <= y22) || (y12 >= y21 && y12 <= y22)) || ((y21 >= y11 && y21 <= y12) || (y22 >= y11 && y22 <= y12)))
            return true;
        else
            return false;
    };
    CanvasSectionContainer.prototype.doSectionsIntersectOnXAxis = function (section1, section2) {
        var x11 = section1.myTopLeft[0];
        var x12 = section1.myTopLeft[0] + section1.size[0];
        var x21 = section2.myTopLeft[0];
        var x22 = section2.myTopLeft[0] + section2.size[0];
        if (((x11 >= x21 && x11 < x22) || (x12 >= x21 && x12 <= x22)) || ((x21 >= x11 && x21 <= x12) || (x22 >= x11 && x22 <= x12)))
            return true;
        else
            return false;
    };
    // Find the left most point from a position with same zIndex.
    CanvasSectionContainer.prototype.hitLeft = function (section) {
        var maxX = -Infinity;
        for (var i = 0; i < this.sections.length; i++) {
            if (this.sections[i].isLocated && this.sections[i].zIndex === section.zIndex && this.sections[i].name !== section.name) {
                var currentLeft = this.sections[i].myTopLeft[0] + this.sections[i].size[0];
                if (currentLeft > maxX && currentLeft < section.myTopLeft[0]) {
                    if (this.doSectionsIntersectOnYAxis(this.sections[i], section)) {
                        maxX = currentLeft;
                    }
                }
            }
        }
        if (maxX === -Infinity)
            return 0; // There is nothing on the left of this section.
        else
            return maxX + Math.round(this.dpiScale); // Don't overlap with the section on the left.
    };
    // Find the right most point from a position with same zIndex.
    CanvasSectionContainer.prototype.hitRight = function (section) {
        var minX = Infinity;
        for (var i = 0; i < this.sections.length; i++) {
            if (this.sections[i].isLocated && this.sections[i].zIndex === section.zIndex && this.sections[i].name !== section.name) {
                var currentRight = this.sections[i].myTopLeft[0];
                if (currentRight < minX && currentRight > section.myTopLeft[0]) {
                    if (this.doSectionsIntersectOnYAxis(this.sections[i], section)) {
                        minX = currentRight;
                    }
                }
            }
        }
        if (minX === Infinity)
            return this.right; // There is nothing on the right of this section.
        else
            return minX - Math.round(this.dpiScale); // Don't overlap with the section on the right.
    };
    // Find the top most point from a position with same zIndex.
    CanvasSectionContainer.prototype.hitTop = function (section) {
        var maxY = -Infinity;
        for (var i = 0; i < this.sections.length; i++) {
            if (this.sections[i].isLocated && this.sections[i].zIndex === section.zIndex && this.sections[i].name !== section.name) {
                var currentTop = this.sections[i].myTopLeft[1] + this.sections[i].size[1];
                if (currentTop > maxY && currentTop < section.myTopLeft[1]) {
                    if (this.doSectionsIntersectOnXAxis(this.sections[i], section)) {
                        maxY = currentTop;
                    }
                }
            }
        }
        if (maxY === -Infinity)
            return 0; // There is nothing on the left of this section.
        else
            return maxY + Math.round(this.dpiScale); // Don't overlap with the section on the top.
    };
    // Find the bottom most point from a position with same zIndex.
    CanvasSectionContainer.prototype.hitBottom = function (section) {
        var minY = Infinity;
        for (var i = 0; i < this.sections.length; i++) {
            if (this.sections[i].isLocated && this.sections[i].zIndex === section.zIndex && this.sections[i].name !== section.name) {
                var currentBottom = this.sections[i].myTopLeft[1];
                if (currentBottom < minY && currentBottom > section.myTopLeft[1]) {
                    if (this.doSectionsIntersectOnXAxis(this.sections[i], section)) {
                        minY = currentBottom;
                    }
                }
            }
        }
        if (minY === Infinity)
            return this.bottom; // There is nothing on the left of this section.
        else
            return minY - Math.round(this.dpiScale); // Don't overlap with the section on the bottom.
    };
    CanvasSectionContainer.prototype.createUpdateDivElements = function () {
        var bcr = this.canvas.getBoundingClientRect();
        for (var i = 0; i < this.sections.length; i++) {
            var section = this.sections[i];
            var element = document.getElementById('test-div-' + this.sections[i].name);
            if (!element) {
                element = document.createElement('div');
                element.id = 'test-div-' + this.sections[i].name;
                document.body.appendChild(element);
            }
            element.style.position = 'fixed';
            element.style.zIndex = '0';
            element.style.left = String(bcr.left + Math.round(section.myTopLeft[0] / this.dpiScale)) + 'px';
            element.style.top = String(bcr.top + Math.round(section.myTopLeft[1] / this.dpiScale)) + 'px';
            element.style.width = String(Math.round(section.size[0] / this.dpiScale)) + 'px';
            element.style.height = String(Math.round(section.size[1] / this.dpiScale)) + 'px';
        }
    };
    CanvasSectionContainer.prototype.reNewAllSections = function (redraw) {
        if (redraw === void 0) { redraw = true; }
        this.orderSections();
        this.locateSections();
        for (var i = 0; i < this.sections.length; i++) {
            this.sections[i].onResize();
        }
        this.applyDrawingOrders();
        if (this.testing)
            this.createUpdateDivElements();
        if (redraw)
            this.drawSections();
    };
    CanvasSectionContainer.prototype.roundPositionAndSize = function (section) {
        section.myTopLeft[0] = Math.round(section.myTopLeft[0]);
        section.myTopLeft[1] = Math.round(section.myTopLeft[1]);
        section.size[0] = Math.round(section.size[0]);
        section.size[1] = Math.round(section.size[1]);
    };
    CanvasSectionContainer.prototype.calculateSectionInitialPosition = function (section, index) {
        if (typeof section.anchor[index] === 'string' || section.anchor[index].length === 1) {
            var anchor = typeof section.anchor[index] === 'string' ? section.anchor[index] : section.anchor[index][0];
            if (index === 0)
                return anchor === 'top' ? section.position[1] : (this.bottom - (section.position[1] + section.size[1]));
            else
                return anchor === 'left' ? section.position[0] : (this.right - (section.position[0] + section.size[0]));
        }
        else {
            // If we are here, it means section's edge(s) will be snapped to another section's edges.
            // Count should always be an odd number. Because last variable will be used as a fallback to canvas's edges (top, bottom, right or left).
            // See anchor explanation on top of this file.
            // Correct example: ["header", "bottom", "top"] => Look for section "header", if found, use its bottom, if not found, use canvas's top.
            if (section.anchor[index].length % 2 === 0) {
                console.error('Section: ' + section.name + '. Wrong anchor definition.');
                return 0;
            }
            else {
                var count = section.anchor[index].length;
                var targetSection = null;
                var targetEdge = null;
                for (var i = 0; i < count - 1; i++) {
                    targetSection = this.getSectionWithName(section.anchor[index][i]);
                    if (targetSection) {
                        targetEdge = section.anchor[index][i + 1];
                        break;
                    }
                }
                if (targetSection) {
                    // So, we have target section, we will use its position. Is it located?
                    if (!targetSection.isLocated) {
                        console.error('Section: ' + section.name + '. Target section for anchor should be located before this section.'
                            + ' It means that target section\'s (if zIndex is the same) processing order should be less or its zIndex should be less than this section.');
                        return 0;
                    }
                    else {
                        if (targetEdge === 'top')
                            return targetSection.myTopLeft[1] - Math.round(this.dpiScale);
                        else if (targetEdge === 'bottom')
                            return targetSection.myTopLeft[1] + targetSection.size[1] + Math.round(this.dpiScale);
                        else if (targetEdge === 'left')
                            return targetSection.myTopLeft[0] - Math.round(this.dpiScale);
                        else if (targetEdge === 'right')
                            return targetSection.myTopLeft[0] + targetSection.size[0] + Math.round(this.dpiScale);
                    }
                }
                else {
                    // No target section is found. Use fallback.
                    var anchor = section.anchor[index][count - 1];
                    if (index === 0)
                        return anchor === 'top' ? section.position[1] : (this.bottom - (section.position[1] + section.size[1]));
                    else
                        return anchor === 'left' ? section.position[0] : (this.right - (section.position[0] + section.size[0]));
                }
            }
        }
    };
    CanvasSectionContainer.prototype.expandSection = function (section) {
        if (this.arrayIncludes(section.expand, 'left')) {
            var initialX = section.myTopLeft[0];
            section.myTopLeft[0] = this.hitLeft(section);
            section.size[0] = initialX - section.myTopLeft[0];
        }
        if (this.arrayIncludes(section.expand, 'right')) {
            section.size[0] = this.hitRight(section) - section.myTopLeft[0];
        }
        if (this.arrayIncludes(section.expand, 'top')) {
            var initialY = section.myTopLeft[1];
            section.myTopLeft[1] = this.hitTop(section);
            section.size[1] = initialY - section.myTopLeft[1];
        }
        if (this.arrayIncludes(section.expand, 'bottom')) {
            section.size[1] = this.hitBottom(section) - section.myTopLeft[1];
        }
    };
    CanvasSectionContainer.prototype.locateSections = function () {
        // Reset some values.
        for (var i = 0; i < this.sections.length; i++) {
            this.sections[i].isLocated = false;
            this.sections[i].myTopLeft = null;
        }
        for (var i = 0; i < this.sections.length; i++) {
            var section = this.sections[i];
            if (!section.boundToSection) {
                section.myTopLeft = [this.calculateSectionInitialPosition(section, 1), this.calculateSectionInitialPosition(section, 0)];
                if (section.expand[0] !== '')
                    this.expandSection(section);
                this.roundPositionAndSize(section);
                section.isLocated = true;
            }
            else {
                var parentSection = this.getSectionWithName(section.boundToSection);
                if (parentSection) {
                    section.myTopLeft = [0, 0];
                    section.size = [0, 0];
                    section.size[0] = parentSection.size[0];
                    section.size[1] = parentSection.size[1];
                    section.myTopLeft[0] = parentSection.myTopLeft[0];
                    section.myTopLeft[1] = parentSection.myTopLeft[1];
                    this.roundPositionAndSize(section);
                    section.isLocated = true;
                }
            }
        }
    };
    CanvasSectionContainer.prototype.orderSections = function () {
        // According to zIndex & processing order.
        for (var i = 0; i < this.sections.length - 1; i++) {
            for (var j = i + 1; j < this.sections.length; j++) {
                if (this.sections[i].zIndex > this.sections[j].zIndex
                    || (this.sections[i].zIndex === this.sections[j].zIndex && this.sections[i].processingOrder > this.sections[j].processingOrder)) {
                    var temp = this.sections[i];
                    this.sections[i] = this.sections[j];
                    this.sections[j] = temp;
                }
            }
        }
    };
    CanvasSectionContainer.prototype.applyDrawingOrders = function () {
        // According to drawing order. Section with the highest drawing order will be drawn on top (inside same zIndex).
        for (var i = 0; i < this.sections.length - 1; i++) {
            var zIndex = this.sections[i].zIndex;
            for (var j = i + 1; j < this.sections.length && this.sections[j].zIndex === zIndex; j++) {
                if (this.sections[i].drawingOrder > this.sections[j].drawingOrder) {
                    var temp = this.sections[j];
                    this.sections[j] = this.sections[i];
                    this.sections[i] = temp;
                }
            }
        }
    };
    CanvasSectionContainer.prototype.drawSectionBorders = function () {
        this.context.lineWidth = 2 * this.dpiScale;
        this.context.strokeStyle = 'blue';
        for (var i = 0; i < this.sections.length; i++) {
            var section = this.sections[i];
            var xStart = section.myTopLeft[0];
            var xEnd = xStart + section.size[0];
            var yStart = section.myTopLeft[1];
            var yEnd = yStart + section.size[1];
            this.context.beginPath();
            this.context.moveTo(xStart, yStart);
            this.context.lineTo(xEnd, yStart);
            this.context.stroke();
            this.context.beginPath();
            this.context.moveTo(xEnd, yStart);
            this.context.lineTo(xEnd, yEnd);
            this.context.stroke();
            this.context.beginPath();
            this.context.moveTo(xEnd, yEnd);
            this.context.lineTo(xStart, yEnd);
            this.context.stroke();
            this.context.beginPath();
            this.context.moveTo(xStart, yEnd);
            this.context.lineTo(xStart, yStart);
            this.context.stroke();
        }
    };
    CanvasSectionContainer.prototype.setPenPosition = function (section) {
        this.context.setTransform(1, 0, 0, 1, 0, 0);
        this.context.translate(section.myTopLeft[0], section.myTopLeft[1]);
    };
    CanvasSectionContainer.prototype.clearNonDocContentArea = function () {
        // Assumes origin is at canvas's (0, 0).
        // Document content area is not necessarily the tile section area
        // or what is represented by map's getPixelBoundsCore().
        this.context.fillStyle = this.clearColor;
        var tileSection = this.getSectionWithName(L.CSections.Tiles.name);
        if (!tileSection) {
            this.context.clearRect(0, 0, this.canvas.width, this.canvas.height);
            this.context.fillRect(0, 0, this.canvas.width, this.canvas.height);
            return;
        }
        // compute topleft and bottom right of "tile content area" in absolute canvas coordinates.
        var contentTopLeft = [
            this.documentTopLeft[0] >= 0 ? tileSection.myTopLeft[0] : tileSection.myTopLeft[0] - this.documentTopLeft[0],
            this.documentTopLeft[1] >= 0 ? tileSection.myTopLeft[1] : tileSection.myTopLeft[1] - this.documentTopLeft[1]
        ];
        var contentBottomRight = [
            tileSection.myTopLeft[0] + this.documentBottomRight[0] - this.documentTopLeft[0],
            tileSection.myTopLeft[1] + this.documentBottomRight[1] - this.documentTopLeft[1]
        ];
        if (this.documentTopLeft[0] < 0)
            contentBottomRight[0] += this.documentTopLeft[0];
        if (this.documentTopLeft[1] < 0)
            contentBottomRight[1] += this.documentTopLeft[1];
        // top margin
        this.context.fillRect(0, 0, this.canvas.width, contentTopLeft[1]);
        // bottom margin
        this.context.fillRect(0, contentBottomRight[1], this.canvas.width, this.canvas.height - contentBottomRight[1]);
        // left margin
        this.context.fillRect(0, contentTopLeft[1], contentTopLeft[0], contentBottomRight[1] - contentTopLeft[1]);
        // right margin
        this.context.fillRect(contentBottomRight[0], contentTopLeft[1], this.canvas.width - contentBottomRight[0], contentBottomRight[1] - contentTopLeft[1]);
    };
    CanvasSectionContainer.prototype.drawSections = function (frameCount, elapsedTime) {
        if (frameCount === void 0) { frameCount = null; }
        if (elapsedTime === void 0) { elapsedTime = null; }
        this.context.setTransform(1, 0, 0, 1, 0, 0);
        if (this.zoomChanged) {
            this.clearNonDocContentArea();
        }
        else {
            this.context.fillStyle = this.clearColor;
            this.context.clearRect(0, 0, this.canvas.width, this.canvas.height);
            this.context.fillRect(0, 0, this.canvas.width, this.canvas.height);
        }
        this.context.font = String(20 * this.dpiScale) + "px Verdana";
        for (var i = 0; i < this.sections.length; i++) {
            if (this.sections[i].isLocated) {
                this.context.translate(this.sections[i].myTopLeft[0], this.sections[i].myTopLeft[1]);
                if (this.sections[i].backgroundColor) {
                    this.context.fillStyle = this.sections[i].backgroundColor;
                    this.context.fillRect(0, 0, this.sections[i].size[0], this.sections[i].size[1]);
                }
                this.sections[i].onDraw(frameCount, elapsedTime);
                if (this.sections[i].borderColor) { // If section's border is set, draw its borders after section's "onDraw" function is called.
                    this.context.lineWidth = this.dpiScale;
                    this.context.strokeStyle = this.sections[i].borderColor;
                    this.context.strokeRect(0.5, 0.5, this.sections[i].size[0], this.sections[i].size[1]);
                }
                this.context.translate(-this.sections[i].myTopLeft[0], -this.sections[i].myTopLeft[1]);
            }
        }
        //this.drawSectionBorders();
    };
    CanvasSectionContainer.prototype.doesSectionExist = function (name) {
        if (name && typeof name === 'string') {
            for (var i = 0; i < this.sections.length; i++) {
                if (this.sections[i].name === name) {
                    return true;
                }
            }
            return false;
        }
        else {
            return false;
        }
    };
    CanvasSectionContainer.prototype.checkNewSectionName = function (options) {
        if (options.name !== undefined && typeof options.name === 'string' && options.name.trim() !== '') {
            if (this.doesSectionExist(options.name)) {
                console.error('There is a section with the same name. Use doesSectionExist for existancy checks.');
                return false;
            }
            else if (this.arrayIncludes(['top', 'left', 'bottom', 'right'], options.name.trim())) {
                console.error('"top", "left", "bottom" and "right" words are reserved. Choose another name for the section.');
                return false;
            }
            else {
                return true;
            }
        }
        else {
            console.error('Sections should have a "name" property.');
            return false;
        }
    };
    CanvasSectionContainer.prototype.checkSectionProperties = function (options) {
        if (options.anchor === undefined
            || options.position === undefined
            || options.size === undefined
            || options.expand === undefined
            || options.processingOrder === undefined
            || options.drawingOrder === undefined
            || options.zIndex === undefined
            || options.interactable === undefined) {
            console.error('Section has missing properties. See "checkSectionProperties" function.');
            return false;
        }
        else {
            return true;
        }
    };
    CanvasSectionContainer.prototype.newSectionChecks = function (options) {
        if (!this.checkNewSectionName(options))
            return false;
        else if (!this.checkSectionProperties(options))
            return false;
        else
            return true;
    };
    CanvasSectionContainer.prototype.createSection = function (options, parentSectionName) {
        if (parentSectionName === void 0) { parentSectionName = null; }
        if (this.newSectionChecks(options)) {
            // Every section can draw from Point(0, 0), their drawings will be translated to myTopLeft position.
            var newSection = new CanvasSectionObject(options);
            newSection.boundToSection = parentSectionName;
            this.pushSection(newSection);
            return true;
        }
        else {
            return false;
        }
    };
    CanvasSectionContainer.prototype.addSection = function (newSection) {
        if (this.newSectionChecks(newSection)) {
            this.pushSection(newSection);
            return true;
        }
        else {
            return false;
        }
    };
    CanvasSectionContainer.prototype.addSectionFunctions = function (section) {
        section.setDrawingOrder = function (drawingOrder) {
            section.drawingOrder = drawingOrder;
            section.containerObject.updateBoundSectionLists();
            section.containerObject.reNewAllSections();
        };
        section.setZIndex = function (zIndex) {
            section.zIndex = zIndex;
            section.containerObject.updateBoundSectionLists();
            section.containerObject.reNewAllSections();
        };
        section.bindToSection = function (sectionName) {
            section.boundToSection = sectionName;
            section.containerObject.updateBoundSectionLists();
            section.containerObject.reNewAllSections();
        };
        section.stopPropagating = function () {
            section.containerObject.lowestPropagatedBoundSection = section.name;
        };
        section.startAnimating = function (options) {
            return section.containerObject.startAnimating(section.name, options);
        };
        section.resetAnimation = function () {
            section.containerObject.resetAnimation(section.name);
        };
    };
    CanvasSectionContainer.prototype.pushSection = function (newSection) {
        // Every section can draw from Point(0, 0), their drawings will be translated to myTopLeft position.
        newSection.context = this.context;
        newSection.documentTopLeft = this.documentTopLeft;
        newSection.containerObject = this;
        newSection.dpiScale = this.dpiScale;
        newSection.sectionProperties.section = newSection;
        this.sections.push(newSection);
        this.addSectionFunctions(newSection);
        this.updateBoundSectionLists();
        newSection.onInitialize();
        this.reNewAllSections(false);
        this.drawSections();
    };
    CanvasSectionContainer.prototype.removeSection = function (name) {
        var found = false;
        for (var i = 0; i < this.sections.length; i++) {
            if (this.sections[i].name === name) {
                var element = document.getElementById('test-div-' + this.sections[i].name);
                if (element) // Remove test div if exists.
                    document.body.removeChild(element);
                this.sections[i].onRemove();
                this.sections[i] = null;
                this.sections.splice(i, 1);
                found = true;
                break;
            }
        }
        if (found) {
            this.updateBoundSectionLists();
            this.reNewAllSections();
            return true;
        }
        else {
            return false;
        }
    };
    CanvasSectionContainer.prototype.setAnimatingSectionName = function (sectionName) {
        this.animatingSectionName = sectionName;
    };
    CanvasSectionContainer.prototype.getAnimatingSectionName = function () {
        return this.animatingSectionName;
    };
    CanvasSectionContainer.prototype.animate = function (timeStamp) {
        if (this.lastFrameStamp > 0)
            this.elapsedTime += timeStamp - this.lastFrameStamp;
        this.lastFrameStamp = timeStamp;
        if (this.duration && this.elapsedTime >= this.duration) { // This is not the only place that can set "continueAnimating" to "false".
            this.continueAnimating = false;
        }
        if (this.continueAnimating) {
            this.drawSections(this.frameCount, this.elapsedTime);
            this.frameCount++;
            requestAnimationFrame(this.animate.bind(this));
        }
        else {
            for (var i = 0; i < this.stoppingFunctionList.length; i++) {
                this.canvas.removeEventListener(this.stoppingEventTypes[i], this.stoppingFunctionList[i], true);
            }
            var section = this.getSectionWithName(this.getAnimatingSectionName());
            if (section) {
                section.isAnimating = false;
                section.onAnimationEnded(this.frameCount, this.elapsedTime);
            }
            this.setAnimatingSectionName(null);
            this.frameCount = this.elapsedTime = null;
            this.drawSections();
        }
    };
    CanvasSectionContainer.prototype.createStoppingFunction = function () {
        var that = this;
        return function () {
            that.continueAnimating = false;
        };
    };
    // Resets animation duration. Not to be called directly. Instead, use (inside section class) this.resetAnimation()
    CanvasSectionContainer.prototype.resetAnimation = function (sectionName) {
        if (sectionName === this.getAnimatingSectionName()) {
            this.lastFrameStamp = 0;
            this.elapsedTime = 0;
            this.frameCount = 0;
        }
    };
    CanvasSectionContainer.prototype.stopAnimating = function () {
        // Though this function is available for every section, generally, only the section that started animation should use this.
        this.continueAnimating = false;
    };
    // Don't call this directly. Instead, call (inside the section class) section.startAnimating(options).
    CanvasSectionContainer.prototype.startAnimating = function (sectionName, options) {
        /*
            Most of the time, we need to draw entire canvas when animating.
            Because if there is another section under the animated one, that section needs to be renewed too.
            Also, sections may need to be redrawn because of the updated view (while animating).
            This animation feature re-draws all sections with requestAnimationFrame.
            Developer can set options to ensure the animation stops at certain point.
            If your section's "onDraw" function is given the variables "frameCount, elapsedTime", then you can assume that CanvasSectionContainer is in animation mode.
            You can also check getAnimatingSectionName function of container class (if it is null or not) to see if animating is on.

            Sections other than the one which started the animation, can't know when the animation will stop.

            The section which started the animation => "(inside section class) this.containerObject.getAnimatingSectionName()".

            For now, only one section can start animations at a time.

            options (possible values are separated by the '|' char):
                // Developer can specify the events those will stop animating.
                // Important note: User shouldn't depend on the order of the events if they assign a stopping handler to an event.
                    Let's assume a stopping function is bound to 'onclick' event.
                    For now, most probably, onclick event will first be propagated to sections, then animation will stop.
                    But when Leaflet is removed, animation will stop first and then onclick event will be propagated to sections.

                stoppingEvents: ['click', 'mousemove' ..etc] // Events should match the real keywords.
                // Developer can set the duration for the animation, in miliseconds. There are also other ways to stop the animation.
                duration: 2000 | null // 2 seconds | null.
        */
        if (!this.getAnimatingSectionName()) {
            this.setAnimatingSectionName(sectionName);
            this.getSectionWithName(sectionName).isAnimating = true;
            this.lastFrameStamp = 0;
            this.continueAnimating = true;
            this.duration = options.duration ? options.duration : null;
            this.elapsedTime = 0;
            this.frameCount = 0;
            this.stoppingFunctionList = new Array(0);
            this.stoppingEventTypes = new Array(0);
            if (options.stoppingEvents) {
                for (var i = 0; i < options.stoppingEvents.length; i++) {
                    this.stoppingEventTypes.push(options.stoppingEvents[i]);
                    this.stoppingFunctionList.push(this.createStoppingFunction());
                    this.canvas.addEventListener(options.stoppingEvents[i], this.stoppingFunctionList[i], true);
                }
            }
            this.animate(performance.now());
            return true;
        }
        else {
            return false;
        }
    };
    return CanvasSectionContainer;
}());
